
From: Dave Hansen <haveblue@us.ibm.com>

discontig.c has some assumptions that mem_map[]s inside of a node are
contiguous.  Teach it to make sure that each region that it's brining online
is actually made up of valid ranges of ram.

Written-by: Andy Whitcroft <apw@shadowen.org>
Signed-off-by: Dave Hansen <haveblue@us.ibm.com>
Signed-off-by: Andrew Morton <akpm@osdl.org>
---

 25-akpm/arch/i386/mm/discontig.c |   14 ++++++++++++++
 25-akpm/arch/i386/mm/init.c      |    2 +-
 25-akpm/include/asm-i386/page.h  |    2 ++
 3 files changed, 17 insertions(+), 1 deletion(-)

diff -puN arch/i386/mm/discontig.c~resubmit-sparsemem-base-teach-discontig-about-sparse-ranges arch/i386/mm/discontig.c
--- 25/arch/i386/mm/discontig.c~resubmit-sparsemem-base-teach-discontig-about-sparse-ranges	2005-04-07 18:32:43.000000000 -0700
+++ 25-akpm/arch/i386/mm/discontig.c	2005-04-07 18:32:43.000000000 -0700
@@ -216,6 +216,7 @@ static unsigned long calculate_numa_rema
 {
 	int nid;
 	unsigned long size, reserve_pages = 0;
+	unsigned long pfn;
 
 	for_each_online_node(nid) {
 		/*
@@ -234,6 +235,19 @@ static unsigned long calculate_numa_rema
 		size = (size + LARGE_PAGE_BYTES - 1) / LARGE_PAGE_BYTES;
 		/* now the roundup is correct, convert to PAGE_SIZE pages */
 		size = size * PTRS_PER_PTE;
+
+		/*
+		 * Validate the region we are allocating only contains valid
+		 * pages.
+		 */
+		for (pfn = node_end_pfn[nid] - size;
+		     pfn < node_end_pfn[nid]; pfn++)
+			if (!page_is_ram(pfn))
+				break;
+
+		if (pfn != node_end_pfn[nid])
+			size = 0;
+
 		printk("Reserving %ld pages of KVA for lmem_map of node %d\n",
 				size, nid);
 		node_remap_size[nid] = size;
diff -puN arch/i386/mm/init.c~resubmit-sparsemem-base-teach-discontig-about-sparse-ranges arch/i386/mm/init.c
--- 25/arch/i386/mm/init.c~resubmit-sparsemem-base-teach-discontig-about-sparse-ranges	2005-04-07 18:32:43.000000000 -0700
+++ 25-akpm/arch/i386/mm/init.c	2005-04-07 18:32:43.000000000 -0700
@@ -191,7 +191,7 @@ static inline int page_kills_ppro(unsign
 
 extern int is_available_memory(efi_memory_desc_t *);
 
-static inline int page_is_ram(unsigned long pagenr)
+int page_is_ram(unsigned long pagenr)
 {
 	int i;
 	unsigned long addr, end;
diff -puN include/asm-i386/page.h~resubmit-sparsemem-base-teach-discontig-about-sparse-ranges include/asm-i386/page.h
--- 25/include/asm-i386/page.h~resubmit-sparsemem-base-teach-discontig-about-sparse-ranges	2005-04-07 18:32:43.000000000 -0700
+++ 25-akpm/include/asm-i386/page.h	2005-04-07 18:32:43.000000000 -0700
@@ -119,6 +119,8 @@ static __inline__ int get_order(unsigned
 
 extern int sysctl_legacy_va_layout;
 
+extern int page_is_ram(unsigned long pagenr);
+
 #endif /* __ASSEMBLY__ */
 
 #ifdef __ASSEMBLY__
_
